def emit_envsetup( oldenv = None ):
    emit_env_dict( calculate_env_setup( oldenv ) )

def emit_env_unsetup( oldenv = None ):
    emit_env_dict( calculate_env_unsetup( oldenv ) )

def create_install_env_clone( env_dict = None ):
    if env_dict is None:
        import os
        env_dict = os.environ.copy()
    else:
        if hasattr( env_dict, 'copy' ):
            env_dict = env_dict.copy()
        else:
            import copy
            env_dict = copy.deepcopy(env_dict)
    env_dict_changes = calculate_env_setup()
    for k,v in env_dict_changes.items():
        if v is None:
            if k in env_dict:
                del env_dict[k]
        else:
            env_dict[k] = v
    return env_dict

def apply_envsetup_to_dict( current_env ):
    for k,v in calculate_env_setup( current_env ).items():
        if v is None:
            if k in current_env:
                del current_env[k]
        else:
            current_env[k] = v

def apply_envunsetup_to_dict( current_env ):
    for k,v in calculate_env_unsetup( current_env ).items():
        if v is None:
            if k in current_env:
                del current_env[k]
        else:
            current_env[k] = v

def calculate_env_unsetup( oldenv = None ):
    #returns dict( var -> value ), with None values meaning the variable should
    #be unset.
    if oldenv is None:
        import os
        oldenv = os.environ
    env_dict = undo_previous_pathvar_changes( oldenv )
    for e in [ 'SBLD_INSTALL_PREFIX','SBLD_DATA_DIR','SBLD_LIB_DIR',
               'SBLD_TESTREF_DIR','SBLD_INCLUDE_DIR',
               '_SIMPLEBUILD_CURRENT_ENV',
              ]:
        if e in env_dict or e in oldenv:
            env_dict[e] = None
    return env_dict

def verify_env_already_setup( oldenv = None ):
    #For unit test
    import os
    if oldenv is None:
        oldenv = os.environ
    e = calculate_env_setup()
    if e:
        print('ERROR: sbld environment not enabled. Requested changes are:')
        for k,v in sorted(e.items()):
            print(f'  {k}={v}')
        raise SystemExit(1)

def calculate_env_setup( oldenv = None ):
    #returns dict( var -> value ), with None values meaning the variable should
    #be unset.

    #First undo effects of any previous setup:
    if oldenv is None:
        import os
        oldenv = os.environ
    env_dict = calculate_env_unsetup( oldenv )

    #Figure out what we need in terms of installdir and env_path variables, so
    #we can inject them with the correct values:
    from .envcfg import var
    instdir = var.install_dir_resolved
    fpcontent=[str(instdir)]
    for pathvar, inst_subdirs in sorted( var.env_paths.items() ):
        if '%' in pathvar:
            from . import error
            error.error(f"Percent sign not allowed in variable name: {pathvar}")
        pathvar_orig_val = env_dict.get(pathvar,oldenv.get(pathvar))
        _prefix = '%' if pathvar_orig_val is None else ''
        fpcontent.append( _prefix + pathvar )
    if any(':' in e for e in fpcontent):
        from . import error
        error.error("Colons not allowed in cache dir or variable names")

    #So inject our new variables:
    for pathvar, inst_subdirs in sorted( var.env_paths.items() ):
        prepend_entries = [ instdir / sd for sd in sorted(inst_subdirs) ]
        ed = env_dict if pathvar in env_dict else oldenv
        env_dict[pathvar] = modify_path_var( pathvar,
                                             env_dict = ed,
                                             blockpath = instdir,
                                             prepend_entries = prepend_entries )


    #Set relevant non-path vars:
    env_dict['_SIMPLEBUILD_CURRENT_ENV'] = ':'.join(str(e) for e in fpcontent)
    env_dict['SBLD_INSTALL_PREFIX']  = str(instdir)
    env_dict['SBLD_DATA_DIR']        = str(instdir/'data')
    env_dict['SBLD_LIB_DIR']         = str(instdir/'lib')
    env_dict['SBLD_TESTREF_DIR']     = str(instdir/'tests'/'testref')
    env_dict['SBLD_INCLUDE_DIR']     = str(instdir/'include')

    #Finally, remove those entries already at a correct value:
    for k in list(_ for _ in env_dict.keys()):
        v = env_dict[k]
        if v is None:
            if k not in oldenv:
                #already not there
                del env_dict[k]
            continue
        else:
            if oldenv.get(k)==v:
                #already at the right value:
                del env_dict[k]
    return env_dict

def undo_previous_pathvar_changes( oldenv ):
    assert oldenv is not None
    env = {}
    oldfp = oldenv.get('_SIMPLEBUILD_CURRENT_ENV')
    if not oldfp:
        return env
    import pathlib
    _ = oldfp.split(':')
    old_instdir, old_pathvars = pathlib.Path(_[0]), set(_[1:])
    for raw_pathvar in old_pathvars:
        if raw_pathvar.startswith('%'):
            was_absent = True
            pathvar = raw_pathvar[1:]
        else:
            was_absent = False
            pathvar = raw_pathvar
        newpv = modify_path_var( pathvar,
                                 env_dict=oldenv,
                                 blockpath=old_instdir)
        if newpv or not was_absent:
            env[pathvar] = newpv or ''
        elif pathvar in oldenv:
            env[pathvar] = None#this means unset
    return env

def emit_env_dict( env_dict):
    import shlex
    from . import io as _io
    for k,v in sorted(env_dict.items()):
        if v is None:
            #always export before unset, since unset statement might be an error
            #if not already set:
            _io.raw_print_ignore_quiet(f'export {k}=')
            _io.raw_print_ignore_quiet(f'unset {k}')
        else:
            _io.raw_print_ignore_quiet('export %s=%s'%(k,shlex.quote(str(v))))

def modify_path_var( varname, *,
                     env_dict,
                     blockpath = None,
                     prepend_entries = None ):
    """Removes all references to blockpath or its subpaths from a path variable,
    prepends any requested paths, and returns the result."""
    import pathlib
    assert env_dict is not None
    assert isinstance(blockpath,pathlib.Path)
    #assert not blockpath.exists() or blockpath.is_dir()#still ok?
    if prepend_entries:
        res = prepend_entries[:]
    else:
        res = []
    from .utils import path_is_relative_to
    for e in (env_dict.get(varname,'') or '').split(':'):
        #NB: Keeping empty entries, to make setup->unsetup more likely to give
        #consistent results:
        if ( blockpath is None
             or not path_is_relative_to( pathlib.Path(e), blockpath ) ):
            res.append(str(e or ''))
    #NB: keeping duplicate entries, to make setup->unsetup more likely to
    #give consistent results:
    #return ':'.join(unique_list(str(e) for e in res))
    return ':'.join(str(e) for e in res)

#def unique_list(seq):
#    seen = set()
#    return [x for x in seq if not (x in seen or seen.add(x))]



import sys
if not ( (3,8) <= sys.version_info[0:2] ):
    from . import error
    error.error('Python not found in required version (3.8+ required).')

def version():
    from ._determine_version import determine_version
    return determine_version()



_cache = [None]
def locate_master_cfg_file():
    if _cache[0] is None:
        _cache[0] = _actual_locate_master_cfg_file()
    return _cache[0]

def _actual_locate_master_cfg_file():
    #Be very conservative about imports here!
    import os
    import pathlib
    def err(msg):
        from . import error
        error.error(msg)
    p = os.environ.get('SIMPLEBUILD_CFG')
    if p:
        p = pathlib.Path(p).expanduser()
        if not p.exists():
            err('SIMPLEBUILD_CFG was set to a non-existing directory or file')
        if p.is_dir():
            p = p / 'simplebuild.cfg'
            if not p.exists():
                err('SIMPLEBUILD_CFG was set to a directory'
                            ' with no simplebuild.cfg file in it')
        else:
            if not p.exists():
                err('SIMPLEBUILD_CFG was set to non-existing file')
        if not p.is_absolute():
            err('SIMPLEBUILD_CFG must be set to an absolute path')
        return p
    p = pathlib.Path('.').absolute()#NB: NOT .resolve() on purpose!
    f = p / 'simplebuild.cfg'
    if f.exists():
        return f
    for p in sorted(p.parents,reverse=True):
        f = p / 'simplebuild.cfg'
        if f.exists():
            return f



def get_sb_argparse_obj():
    return parse_args(argv=['sb'],
                      sphinx_mode=True)

def parse_args( argv = None,
                return_parser = False,
                sphinx_mode = False ):

    import argparse
    import sys
    import os
    import pathlib

    if argv is None:
        argv = sys.argv[:]
    progname = os.path.basename(argv[0])


    descr="""

Invokes the simplebuild configuration and build system. Usually this means
locating a simplebuild.cfg file in the current (or a parent directory), and
using the corresponding configuration. The default action is to configure and
build those packages, but as described below, certain options might change this
in order to instead display information, launch unit tests, prepare new
projects, etc.


"""

    #Special for --help+--init mode:
    initmode_shorthelp=('Initialise a new simplebuild bundle in the current\n'
                        'directory by creating a simplebuild.cfg file. Use\n'
                        '--init --help for more options.')
    _p = progname
    initmode_longhelp = f"""

New project initialisation mode ("{_p} --init") is used to initialise a new
simplebuild bundle in the current directory by creating a simplebuild.cfg
file. Additional arguments can be used to specify bundles which the new project
should depend on, or they can be a special keyword (DEBUG|COMPACT|...). In any
case, it might be most convenient to edit the simplebuild.cfg file afterwards to
fine-tune the desired settings.

Examples of {_p} --init usage:

$> {_p} --init core_val

    Set up project which depends on the core_val project.

$> {_p} --init DEBUG

    Set up project with build options for producing debug symbols
    (build.mode='debug').

$> {_p} --init COMPACT

    The resulting simplebuild.cfg file will have all comments and empty lines
    stripped.

$> {_p} --init core_val dgcode COMPACT DEBUG

    Set up project with dependency on core_val and dgcode bundles, with the
    build option for producing debug symbols, and a compact simplebuild.cfg
    file.

$> {_p} --init dgcode CACHEDIR::/some/where

    Set up project with a dependency on the dgcode bundle, and the
    build.cachedir option set to the directory "/some/where".

    """.strip()+'\n'

    if '--init' in argv and ('-h' in argv or '--help' in argv):
        print(initmode_longhelp)
        raise SystemExit

    parser = argparse.ArgumentParser(
        usage=f'{progname} [-h|--help] [options]',
        description=descr.strip(),
        formatter_class=argparse.RawTextHelpFormatter)

    exclusive = set()

    group_build = parser.add_argument_group(
        'Controlling the build',
        "The build will be carried out and installed in a cache directory "
        "(use -s to\nsee which). The following options affects the build "
        "process.")

    group_qv = group_build.add_mutually_exclusive_group()
    group_qv.add_argument("-v", "--verbose", action="store_true",
                          help="Enable more verbosity.")
    group_qv.add_argument("-q", "--quiet", action="store_true",
                          help='Less verbose. Silence even some warnings.')
    group_build.add_argument(
        "-j", "--jobs",
        type=int, dest="njobs", default=0, metavar="N",
        help="Use up to N parallel processes (default: auto).")
    group_build.add_argument(
        "-e", "--examine",
        action='store_true', dest="examine",
        help="Force (re)examination of environment.")
    group_build.add_argument(
        "-i", "--insist",
        action="store_true", dest="insist",
        help="Insist on reconf/rebuild/reinstall from scratch.")

    group_query_g = parser.add_argument_group('Query options')
    group_query = group_query_g.add_mutually_exclusive_group()

    group_query.add_argument("--version",
                             action="store_true", dest="show_version",
                             help="Show simplebuild version.")
    exclusive.add(('--version','show_version'))

    group_query.add_argument('-s','--summary',
                             action='store_true', dest='summary',
                             help='Print a summary of the configuration.')
    exclusive.add('summary')

    group_query.add_argument("--pkginfo",
                             action="store", dest="pkginfo", default='',metavar='PKG',
                             help="Print information about package PKG.")
    exclusive.add('pkginfo')
    group_query.add_argument("--incinfo",action="store", dest="incinfo",
                             default='',metavar='CFILE',
                             help=( "Show inclusion relationships for CFILE in"
                                    " C++/C\nformat. Specify multiple files with"
                                    " comma-\nseparation and wildcards (\"*\").") )
    exclusive.add('incinfo')
    group_query.add_argument("--pkggraph",
                             action="store_true", dest="pkggraph",
                             help="Create graph of package dependencies.")
    exclusive.add('pkggraph')

    group_query.add_argument("--activegraph",
                             action="store_true", dest="pkggraph_activeonly",
                             help="Create graph for enabled packages only.")
    exclusive.add(('--activegraph','pkggraph_activeonly'))

    group_query.add_argument("--grep",action="store", dest="grep",
                             default='',metavar='PATTERN',
                             help="Grep files in packages for PATTERN.")
    exclusive.add('grep')
    group_query.add_argument("--grepc",action="store", dest="grepc",
                             default='',metavar='PATTERN',
                             help="Like --grep but show only count per package.")
    exclusive.add('grepc')
    group_query.add_argument("--find",action="store", dest="find",
                             default=None, metavar='PATTERN',
                             help=("Search for file and path names"
                                   " matching PATTERN."))
    exclusive.add('find')

    #NB: Do not change the title in title in the next line without updating its
    #usage as anchor name in the sphinx documentation:
    group_init = parser.add_argument_group('New project initialisation options')
    group_init.add_argument('--init', action='store_true', dest='init',
                            help=(initmode_longhelp
                                  if sphinx_mode else initmode_shorthelp ))
    exclusive.add('init')

    group_test = parser.add_argument_group('Unit testing options')

    group_test.add_argument("-t", "--tests",
                           action="store_true", dest="runtests",
                           help="Run unit tests after the build.")
    exclusive.add(('--tests','runtests'))
    group_test.add_argument("--testexcerpts", type=int, dest="nexcerpts",
                            default=0,metavar="N",
                            help=("Print N first+last lines of failed test log"
                                  " files."))
    group_test.add_argument("--testfilter", type=str, dest="testfilter",
                            default='', metavar="PATTERN",
                            help=('Only run tests with names matching at least '
                                  'one\nof the provided patterns (use wildcards,'
                                  ' comma\nseparation, and "!" prefix to'
                                  ' negate).'),)

    group_other = parser.add_argument_group('Other options')


    group_other.add_argument("-c", "--clean",
                             action='store_true', dest="clean",
                             help=("Completely remove cache directories and"
                                   " exit."))
    exclusive.add('clean')

    group_other.add_argument("--replace", action="store", dest="replace",
                             default=None, metavar='PATTERN',
                             help=("Global search and replace in packages via"
                                   " pattern\nlike '/OLDCONT/NEWCONT/' (use "
                                   "with care!)."))
    exclusive.add('replace')

    group_other.add_argument("--requirepkg", nargs = '+', dest="requirepkgs",
                             default=None, metavar='PkgName',
                             help=("Fail unless all listed"
                                   " packages are enabled."))

    group_other.add_argument("--removelock",
                           action='store_true', dest="removelock",
                           help="Force removal of lockfile.")

    group_other.add_argument("--env-setup",
                             action="store_true", dest="env_setup",
                             help=("Emit shell code needed to modify"
                                   " environment\nvariables to use"
                                   " build packages, then exit."))
    exclusive.add(('--env-setup','env_setup'))

    a=group_other.add_argument("--env-unsetup",
                             action="store_true", dest="env_unsetup",
                             help=("Emit shell code undoing the effect of"
                                   " any\nprevious --env-setup usage, then exit."))
    exclusive.add(('--env-unsetup','env_unsetup'))

    if sphinx_mode:
        return parser
    args, args_unused = parser.parse_known_args(argv[1:])
    bad_unused=list(a for a in args_unused if a.startswith('-'))
    if bad_unused:
        parser.error('unknown argument: %s'%bad_unused[0])

    present_exclusive = set()
    for a in exclusive:
        if isinstance(a,str):
            flag,attr='--%s'%a,a
        else:
            flag,attr=a
        if bool(getattr(args,attr)):
            present_exclusive.add(flag)
    if len(present_exclusive)>1:
        f1=list(present_exclusive)[0]
        f2=list(present_exclusive)[1]
        parser.error(f'arguments {f1} and {f2} are mutually exclusive')

    if args.grepc:
        args.grep=args.grepc
        args.grepc=bool(args.grepc)

    if args.pkggraph_activeonly:
        args.pkggraph=True

    _ = []
    for e in (args.requirepkgs or []):
        for ee in e.strip().split(','):
            if ee.strip():
                _.append(ee.strip())
    args.requirepkgs = _

    #Some query option require that we always load all package meta
    #data. Additionally, some have "pathzoom", and use the unused positional
    #arguments as path filters:
    query_mode_withpathzoom_n = sum(int(bool(a)) for a in [args.grep,
                                                           args.replace,
                                                           args.find])
    query_mode_n = query_mode_withpathzoom_n + sum(int(bool(a))
                                                   for a in
                                                   [args.pkggraph,
                                                    args.pkginfo,args.
                                                    incinfo,args.summary])
    if int(args.clean)+ query_mode_n > 1:
        parser.error("Conflicting options (+internal inconsistency:"
                     " this conflict should have been caught earlier)")
    args.query_mode = query_mode_n > 0

    args.querypaths=[]

    if query_mode_withpathzoom_n > 0:
        for a in args_unused:
            qp = pathlib.Path(a).expanduser()
            if not qp.exists():
                parser.error(f'Path not found: {qp}')
            qp = qp.absolute().resolve()
            args.querypaths.append(qp)
        args_unused=[]

    if args.init:
        from .singlecfg import is_valid_bundle_name
        args.init = []
        for a in args_unused:
            #All-uppcase option keywords like COMPACT/DEBUG/... are actually
            #valid bundle names, so we don't have to treat them specially
            #here. But we do not to allow for 'CACHEDIR::/some/where' arguments:
            special_init_opt = a.startswith('CACHEDIR::')
            if not special_init_opt and not is_valid_bundle_name(a):
                parser.error(f'Invalid name for dependency bundle: {a}')
            args.init.append( a )
        args_unused = []
    else:
        args.init = None

    if args_unused:
        parser.error("Unrecognised arguments: %s"%' '.join(args_unused))

    if return_parser:
        return parser, args
    else:
        return args



def parse_stdouterr(fh):
    #Errors causes ec!=0 so are noticed by the user. But we want to capture
    #warnings so we can point them out later, and in particular this one:
    #
    #CMake Warning:
    #  Manually-specified variables were not used by the project:
    #
    #    BLA
    #    LALA
    #
    #
    n_warnings=0
    unused_var_mode=False
    unused_vars=[]
    for l in fh:
        l=l.rstrip()
        if not l:
            continue
        if l.startswith('CMake Warning:'):
            n_warnings += 1
            continue
        if n_warnings and 'Manually-specified variables were not used by the project:' in l:
            unused_var_mode = True
            continue
        if unused_var_mode:
            if l.startswith('  '):
                l=l.strip()
                assert l
                if l not in unused_vars:
                   unused_vars.append( l )
                continue
            else:
                assert unused_vars
                unused_var_mode=False
    assert bool(unused_vars)==bool(n_warnings)
    return {'other_warnings':n_warnings-(1 if unused_vars else 0),'unused_vars':unused_vars}

def parse(fh):
    extdeps = {}
    sysvars = {'langs':{}}
    cmakevars = {}
    for l in fh:
        c=l[0]
        if c=='#':
            continue
        l=[e.strip() for e in l.split('@',2)]
        if l[0]=='EXT':
            l[2:]=l[2].split('@',1)#can actually have 4 fields in this case, not just 3
            ext=l[1]
            if ext not in extdeps:
                extdeps[ext]={}
            if l[2]=='PRESENT':
                extdeps[ext]['present']=bool(int(l[3]))
            elif l[2]=='LINK':
                extdeps[ext]['ldflags']=l[3]
            #elif l[2]=='COMPILE':
            #    extdeps[ext]['cflags']=l[3]
            elif l[2]=='VERSION':
                extdeps[ext]['version']=l[3]
            elif l[2]=='COMPILE_CXX':
                extdeps[ext]['cflags_cxx']=l[3]
            elif l[2]=='COMPILE_C':
                extdeps[ext]['cflags_c']=l[3]
            else:
                assert False,"unexpected field: %s"%l[2]
        elif l[0]=='VAR':
            cmakevars[l[1]]=l[2]
        else:
            assert False



    sysvars['general'] = {
        'cmake_version': cmakevars['CMAKE_VERSION'],
        'system': cmakevars['CMAKE_SYSTEM'],
        'pybind11_embed_cflags_list' : cmakevars['PYBIND11_EMBED_CFLAGS_LIST'],
        'pybind11_embed_linkflags_list' : cmakevars['PYBIND11_EMBED_LINKFLAGS_LIST'],
        'pybind11_module_cflags_list' : cmakevars['PYBIND11_MODULE_CFLAGS_LIST'],
        'pybind11_module_linkflags_list' : cmakevars['PYBIND11_MODULE_LINKFLAGS_LIST'],
        'pybind11_version' : cmakevars['PYBIND11_VERSION'],
    }

    compiler_supports_rpathlink_flag = cmakevars['CAN_USE_RPATHLINK_FLAG']=='1'
    for lang in ['CXX','C','Fortran']:
        lvars={}
        if lang!='Fortran' or ('Fortran' in extdeps and extdeps['Fortran']['present']):
            lvars['cased'] = lang
            lvars['compiler'] = cmakevars['CMAKE_%s_COMPILER'%lang]
            lvars['compiler_version_long'] = cmakevars['CMAKE_%s_COMPILER_VERSION_LONG'%lang]
            lvars['compiler_version_short'] = cmakevars['CMAKE_%s_COMPILER_VERSION_SHORT'%lang]
            lvars['dep_versions'] = cmakevars['SBLD_GLOBAL_VERSION_DEPS_%s'%lang]
            cflags = ' '.join([cmakevars['CMAKE_%s_FLAGS'%lang],cmakevars.get('SBLD_GLOBAL_COMPILE_FLAGS_%s'%lang,'')])
            if lang=='C':
                cflags = ' '.join([c for c in cflags.split() if not (c.startswith('-std=c++') or c.startswith('-std=gnu++'))])
            elif lang=='CXX':
                cflags = ' '.join([c for c in cflags.split() if not (c.startswith('-std=c') and not c.startswith('-std=c++'))])
            lvars['cflags'] = cflags
            lvars['ldflags_prepend'] = cmakevars.get('SBLD_GLOBAL_LINK_FLAGS_PREPENDED','')
            lvars['ldflags'] = ' '.join([cmakevars['CMAKE_%s_LINK_FLAGS'%lang],cmakevars.get('SBLD_GLOBAL_LINK_FLAGS','')])
            lvars['name_obj'] = '%s'+cmakevars['CMAKE_%s_OUTPUT_EXTENSION'%lang]
            lvars['name_lib'] = cmakevars['CMAKE_SHARED_LIBRARY_PREFIX']+'%s'+cmakevars['CMAKE_SHARED_LIBRARY_SUFFIX']
            lvars['name_exe'] = '%s'

            cf_shlib = '${CFLAGSLANG_SHLIB_%s}'%lang.lower()
            cf_exe   = '${CFLAGSLANG_EXE_%s}'%lang.lower()
            lf='${LDFLAGSLANG_%s}'%lang.lower()
            lfprepend='${LDFLAGSPREPENDLANG_%s}'%lang.lower()

            lvars['create_obj_for_shlib'] = cmakevars['RULE_%s_COMPOBJ'%lang].replace('[FLAGS]','%s %%s'%cf_shlib).replace('[INPUT]','%s').replace('[OUTPUT]','%s')
            lvars['create_obj_for_exe']   = cmakevars['RULE_%s_COMPOBJ'%lang].replace('[FLAGS]','%s %%s'%cf_exe).replace('[INPUT]','%s').replace('[OUTPUT]','%s')
            lvars['create_lib'] = cmakevars['RULE_%s_SHLIB'%lang].replace('[FLAGS]','%s %s %s %%s'%(lfprepend,cf_shlib,lf)).replace('[INPUT]','%s').replace('[OUTPUT]','%s')
            lvars['create_exe'] = cmakevars['RULE_%s_EXECUTABLE'%lang].replace('[FLAGS]','%s %s %s %%s'%(lfprepend,cf_exe,lf)).replace('[INPUT]','%s').replace('[OUTPUT]','%s')

            lvars['rpath_flag_lib'] = cmakevars['RPATH_FLAG_%s_SHLIB'%lang]
            lvars['rpath_flag_exe'] = cmakevars['RPATH_FLAG_%s_EXECUTABLE'%lang]
            if lvars['rpath_flag_lib']:
                lvars['rpath_flag_lib'] += '%s'
            if lvars['rpath_flag_exe']:
                lvars['rpath_flag_exe'] += '%s'
            lvars['can_use_rpathlink_flag'] = compiler_supports_rpathlink_flag
        sysvars['langs'][lang.lower()]=lvars

    #copy parts of compilation setup into special volatile section to
    #determine if makefiles need to be rebuild (flags will belong in the
    #main Makefile and so changes in those will not need new package
    #makefiles):
    non_volatile=set(['cflags','ldflags','ldflags_prepend','compiler_version_long','compiler_version_short','dep_versions'])
    sysvars['volatile'] = dict((lang,dict((k,v) for k,v in info.items() if k not in non_volatile)) for lang,info in sysvars['langs'].items())
    sysvars['volatile']['sysgeneral'] = dict( (k,v) for k,v in sorted(sysvars['general'].items()) )

    sysvars['runtime']={'extra_lib_path':[e for e in cmakevars['SBLD_EXTRA_LDLIBPATHS'].split(';') if e],
                        'extra_bin_path':[e for e in cmakevars['SBLD_EXTRA_PATHS'].split(';') if e],
                        'libs_to_symlink':[e for e in cmakevars['SBLD_LIBS_TO_SYMLINK'].split(';') if e]}

    autoreconf={}
    autoreconf['bin_list'] = cmakevars['autoreconf_bin_list']
    autoreconf['env_list'] = cmakevars['autoreconf_env_list']
    cfgvars={'system':sysvars,'extdeps':extdeps,'autoreconf':autoreconf}

    return cfgvars

def extractenv(*,cmakeargs,actually_needed_extdeps,quiet=True,verbose=False):
    from . import io as _io
    print = _io.print
    from . import dirs
    from . import utils
    from . import envcfg
    import shlex

    print("Inspecting environment via CMake")

    cmakeblddir = dirs.blddir / 'cmake'
    cmakesrcdir = dirs.cmakedetectdir
    try:
        utils.rm_rf( cmakeblddir )
    except OSError:
        print(f'Failed to remove temporary cmake bld dir: {cmakeblddir}')
        return
    cmakeblddir.mkdir(parents=True)

    cmd = [ 'cmake', '-S', cmakesrcdir, '-B', cmakeblddir ]
    #Pass on options like build type and conda-forge CMAKE_ARGS:
    cmd += envcfg.var.cmake_args
    #Legacy options (unused now?):
    for a in cmakeargs:
        cmd.append('-D%s'%a)
    assert not (quiet and verbose)
    if quiet:
        cmd.append('-DSBLD_QUIET=1')
    if verbose:
        cmd.append('-DSBLD_VERBOSE=1')
    _joined_extdeps=':'.join(actually_needed_extdeps)
    cmd.append( f'-DSBLD_ACTUALLY_USED_EXTDEPS={_joined_extdeps}' )

    output_file = cmakeblddir / 'cmake_output21_capture.txt'
    output_file_quoted = shlex.quote(str(output_file))
    cmd = ' '.join(shlex.quote(str(e)) for e in cmd )
    if _io.is_quiet():
        cmd += ' >& %s; exit ${PIPESTATUS[0]}'%output_file_quoted
    else:
        cmd += ' 2>&1|tee %s; exit ${PIPESTATUS[0]}'%output_file_quoted

    import time
    t0 = time.time()
    ec = utils.system(cmd)
    t1 = time.time()
    print("Environment inspection done (%.2g seconds)"%(t1-t0))
    if ec!=0:
        return
    with output_file.open('rt') as fh:
        printedinfo = parse_stdouterr( fh )
    with ( cmakeblddir / 'cmakecfg.txt' ).open('rt') as fh:
        writteninfo = parse( fh )
    writteninfo['cmake_printinfo'] = printedinfo
    return writteninfo



import os
import pathlib
from . import conf

_islink=os.path.islink
_osstat=os.stat
def mtime(f):
    s = f.stat() if hasattr(f,'stat') else _osstat(f)
    return max(s.st_mtime,s.st_ctime)
    #m=max(s.st_mtime,s.st_ctime)
    #if not _islink(f):
    #    return m
    #s=os.lstat(f)
    #return max(m,s.st_mtime,s.st_ctime)

def mtime_pathlist(files_and_dirs,depth=0):
    """common (highest) modification time of input files and directories. Directories are searched recursively."""
    join  = os.path.join
    isdir = os.path.isdir
    exists = os.path.exists
    ld = os.listdir
    mt = float('-inf')
    moredirs=[]
    for fd in files_and_dirs:
        if not exists(fd):
            raise RuntimeError('Indicated path does not exist: %s'%fd)
        mt = max(mt,mtime(fd))
        if isdir(fd):
            for f in ld(fd):
                f = join(fd,f)
                if isdir(f):
                    moredirs+=[f]
                else:
                    mt = max(mt,mtime(f))
    if moredirs:
        if depth>20:
            raise RuntimeError('Max recursion depth exceeded in mtime_pathlist')
        mt = max(mt,mtime_pathlist(moredirs,depth+1))
    return mt

def mtime_pkg(pkg):
    #assuming the vast majority of files to be in proper subdirs, we simply take
    #the highest mtime of the package cfg file and any file in any subdir. To detect
    #subdirs disappearing we also take the mtime of the pkg dir itself and the first level of subdirs.
    join  = os.path.join
    isdir = os.path.isdir
    ld = os.listdir
    d=pkg.dirname
    #ctime for metadata, mtime for content
    f=join(d,conf.package_cfg_file)
    mt=max(mtime(f),mtime(d))
    for fd in ld(d):
        if conf.ignore_file(fd):
            continue
        fd=join(d,fd)
        if isdir(fd):
            mt = max(mt,mtime(fd))
            for f in ld(fd):
                if conf.ignore_file(f):
                    continue
                f = join(fd,f)
                mt = max(mt,mtime(f))
    return mt

def mtime_cmake():
    from . import dirs
    files  = list(dirs.cmakedetectdir.glob('**/*.txt'))
    files += list(dirs.cmakedetectdir.glob('**/*.cmake'))
    files += list(dirs.cmakedetectdir.glob('**/*.py'))
    files = [ f for f in files if '#' not in f.name ]
    mt=-1
    for f in files:
        mt=max(mt,mtime(f))
    return mt

def mtime_pymods():
    files = [ str(e) for e in pathlib.Path(__file__).parent.glob('*.py') ]

    files = [ f for f in files if '#' not in f ]
    mt=-1
    for f in files:
        mt=max(mt,mtime(f))
    return mt



#Control all print operations, to guarantee nothing untowards gets printed if
#e.g. running --quiet or --env-setup.:

print_prefix_name = 'sbld'
print_prefix = f'{print_prefix_name}: '
_quiet = [False]#might be overridden by a call to make_quiet
_verbose = [False]#might be overridden by a call to make_verbose/make_quiet
_orig_printfct = print
_printfct = [_orig_printfct]
def print( *args, **kwargs ):
    return _printfct[0]( print_prefix, *args, **kwargs )
def print_no_prefix( *args, **kwargs ):
    return _printfct[0]( *args, **kwargs )
def raw_print_ignore_quiet( *args, **kwargs ):
    return _orig_printfct( *args, **kwargs )
def make_quiet():
    if not _quiet[0]:
        _quiet[0] = True
        _printfct[0] = lambda *args, **kwargs : None
def make_verbose():
    _verbose[0] = True
def is_quiet():
    return _quiet[0]
def is_verbose():
    return _verbose[0] and not _quiet[0]



import os
pjoin=os.path.join

def grepfile(filename,pattern,color=None):
    from . import col
    fh=None
    if filename.endswith('.gz'):
        import gzip
        try:
            fh=gzip.open(filename,'rt')
            #read a bit, to trigger error here (in case file with extension .gz
            #is not actually compressed with gzip)
            try:
                for l in fh:
                    break
            except UnicodeDecodeError:
                return#ignore gzipped non-text data files
            fh.seek(0)#rewind
        except (IOError,OSError):
            fh=None
    if fh is None:
        fh=open(filename)
    lp=pattern.lower()
    try:
        for i,l in enumerate(fh):
            ll=l.lower()
            if lp.lower() in ll:
                if color is None:
                    yield (i+1,l)
                else:
                    used=0
                    l2=''
                    while used<len(ll):
                        k=ll[used:].find(lp)
                        if k==-1:
                            l2+=l[used:]
                            break
                        else:
                            l2+=l[used:used+k]
                            l2+=color
                            l2+=l[used+k:used+k+len(lp)]
                            l2+=col.grep_unmatch
                            used+=k+len(pattern)
                    yield (i+1,col.grep_unmatch+l2+col.end)
    except UnicodeDecodeError:
        pass#ignore non-text data files
    fh.close()

#TODO to dirs.py ?:
def pkgdir_for_search(pkg,*subpath):
    return os.path.join(pkg.dirname,*subpath)

def pkgfiles(pkg):
    #Attempt to return files which are meaningful (to keep it simple, we say it
    #is all files in subdirs + the pkg.info file).
    from . import conf
    d = pkgdir_for_search(pkg)
    yield conf.package_cfg_file
    for d2rel in os.listdir(d):
        d2=pjoin(d,d2rel)
        if os.path.isdir(d2):
            for frel in os.listdir(d2):
                if not conf.ignore_file(frel):
                    if not os.path.isdir(pjoin(d2,frel)):
                        yield pjoin(d2rel,frel)

def grep(pkg,pattern,filenames=None,countonly=False):
    import pathlib
    from . import col
    n=0
    from .io import print
    col_match = None if countonly else col.grep_match
    for f in pkgfiles(pkg):
        f = pathlib.Path(f)
        if filenames and f not in filenames:
            continue
        pdir=pkgdir_for_search(pkg)
        ff=pjoin(pdir,f)
        for linenum,line in grepfile( ff, pattern, color = col_match):
            n+=1
            if countonly:
                continue
            pkgdir=os.path.relpath(pdir)
            pkgdir=(col.ok if pkg.enabled else col.bad) + pkgdir + col.end
            output = '  %s/%s [L%i]: %s'%(pkgdir,f,linenum,line)
            if '\n' not in output:
                output += '\n'
            print(output,end='')
    if countonly and n:
        pn=pkg.name
        pn=(col.ok if pkg.enabled else col.bad) + pn + col.end
        print('%s : %i hits'%(pn.rjust(30),n))
    return n



# Error/warning functions and classes to distinguish exceptions thrown by our
# code from other exceptions.
#
# Typical intended usage:
#
# from . import error
# error.error('some error message')
# error.warn('some warning message')
#
# IMPORTANT: This module should NOT import other simplebuild modules (in
# particular not anything triggering envcfg etc.)

import warnings

class Error(Exception):
    pass

class SimpleBuildUserWarning( UserWarning ):
    """UserWarning's emitted from simplebuild"""
    def __init__(self,*args,**kwargs):
        super(SimpleBuildUserWarning, self).__init__(*args,**kwargs)

default_error_type = SystemExit

def error(*args):
    raise default_error_type('ERROR: '+'\n'.join(args))

#Used to exit the programme after catching and dealing with an Error at a lower level.
class CleanExit(Exception):
    pass
def clean_exit(ec):
    ce=CleanExit()
    ce._the_ec=ec
    raise ce

def print_traceback(exc):
    from .io import print
    print()
    import traceback
    print("----begin traceback---")
    if hasattr(exc,'__traceback__'):
        traceback.print_exception(type(exc),exc,exc.__traceback__)
    else:
        #This can only happen in python2:
        traceback.print_exc(exc)
    print("----end traceback---")
    print()


def direct_print_warning( msg ):
    from .io import print_prefix,print_prefix_name,print_no_prefix
    s = print_prefix.replace(print_prefix_name, '%s WARNING'%print_prefix_name)
    print_no_prefix('%s%s'%(s,msg))

_orig_showwarning = warnings.showwarning
def _custom_warning_fmt(msg,cat,*args,**kwargs):
    if issubclass(cat,SimpleBuildUserWarning):
        direct_print_warning( msg )
    else:
        _orig_showwarning(msg,cat,*args,**kwargs)

def fmt_simplebuild_warnings():
    warnings.showwarning = _custom_warning_fmt

def warn(msg):
    warnings.warn( SimpleBuildUserWarning( str(msg) ), stacklevel = 2 )



def init_project( args = None ):
    args = args[:] if args else []

    def has_keyword( args, kw ):
        _=[e for e in args if e!=kw]
        return _, len(args)!=len(_)

    def extract_opt_with_args( args, optname, pick_last ):
        key=f'{optname}::'
        optvals = []
        otherargs = []
        for e in args:
            if e.startswith(key):
                optvals.append( e[len(key):] )
            else:
                otherargs.append( e )
        if pick_last:
            optvals = optvals[-1] if optvals else None
        return otherargs, optvals

    #Keywords:
    args, debug_mode = has_keyword(args, 'DEBUG')
    args, compact = has_keyword(args, 'COMPACT')
    args, build_cachedir = extract_opt_with_args( args, 'CACHEDIR',
                                                  pick_last = True)

    #Remaining args are the dep-bundles, but remove duplicates:
    depbundles = []
    for d in args:
        if d not in depbundles:
            depbundles.append(d)

    from .io import print
    import pathlib
    cwd = pathlib.Path.cwd()
    template = ( pathlib.Path(__file__).parent
                 / 'data' / 'cfgtemplate.txt' ).read_text()
    res = ''
    for e in template.splitlines():
        res += e.rstrip()+'\n'
        if e.startswith('[build]'):
            if debug_mode:
                #inject mode statement:
                sbundles = "', '".join(depbundles)
                res += "\n  mode = 'debug'\n\n"
            if build_cachedir:
                #inject cachedir statement:
                res += f"\n  cachedir = '{build_cachedir}'\n\n"
        elif e.startswith('[depend]'):
            if depbundles:
                #inject depend.projects list:
                sbundles = "', '".join(depbundles)
                res += f"\n  projects = ['{sbundles}']\n\n"

    if compact:
        res2 = ''
        for e in res.splitlines():
            if '#' in e:
                e = e.split('#',1)[0]
            e = e.rstrip()
            if e:
                res2 += e + '\n'
        res = res2

    for f in cwd.glob('**/simplebuild*.cfg'):
        from .error import error
        error('Can not initialise simplebuild package bundle'
              f' here due to conflicting file: {f}')

    outfile = cwd / 'simplebuild.cfg'
    outfile.write_text(res)
    print(f'Created {outfile.name}. If you wish you can edit it to'
          ' fine-tune your configuration')



import Core.System

def get_n_cores(prefix=''):
    import os
    if os.path.exists('/proc/cpuinfo'):
        #linux
        n=0
        for line in open('/proc/cpuinfo'):
            if ( line.startswith('processor')
                 and line.split()[0:2]==['processor',':'] ):
                n+=1
        if not n:
            print(f'{prefix}Warning: Could not determine number of processors.'
                  ' Assuming just one present (override with -jN)')
            return 1
        return n
    else:
        #osx
        (ec,n)=Core.System.system('sysctl -n hw.ncpu',catch_output=True)
        if ec:
            print(f'{prefix}Warning: Could not determine number of processors.'
                  ' Assuming just one present (override with -jN)')
            return 1
        return int(n.strip())

def get_load(ncores,prefix=''):
    (ec,p)=Core.System.system('/bin/ps -eo pcpu',catch_output=True)
    if ec:
        print(f'{prefix}Warning: Could not determine current CPU load.'
              ' Assuming 0%.')
        return 0.0
    p=0.01*sum([float(x.replace(b',',b'.')) for x in p.split()[1:]])
    if p>ncores*1.5:
        print(f'{prefix}Warning: Could not determine current CPU load'
              ' ("ps -eo pcpu" output was suspicous). Assuming 0%.')
        return 0.0
    return p

def auto_njobs(prefix = ''):
    nc=get_n_cores(prefix)
    freecores=nc-get_load(nc,prefix)
    return max(1,int(0.5+0.98*freecores))




def _handle_env_setup( args ):
    def _check_nargs():
        if len(args)>1:
            raise SystemExit('Error: Do not use --env-setup or'
                             ' --env-unsetup with other arguments!')
    if any(a.startswith('--env-u') for a in args):
        _check_nargs()
        #Enable --env-unsetup usage, even outside a simplebuild project:
        from . import io as _io
        _io.make_quiet()
        from .envsetup import emit_env_unsetup
        emit_env_unsetup()
        raise SystemExit
    if any(a.startswith('--env-s') for a in args):
        _check_nargs()
        #Short-circuit to efficiently enable --env-setup call:
        from . import io as _io
        _io.make_quiet()
        from .envsetup import emit_envsetup
        emit_envsetup()
        raise SystemExit

def main( prevent_env_setup_msg = False ):
    import sys
    _args = sys.argv[1:]
    if any( e.startswith('--env-') for e in _args ):
        #Special short-circuit for --env-[un]setup:
        _handle_env_setup(_args)
    from . import frontend
    frontend.simplebuild_main( prevent_env_setup_msg = prevent_env_setup_msg )

def unwrapped_main():
    #For the unwrapped_simplebuild entry point, presumably only called from a
    #shell function which might provide a special --env-setup-silent-fail.
    import sys
    if len(sys.argv)==2 and sys.argv[1]=='--env-setup-silent-fail':
        #Quietly check if we are in a simplebuild project, and abort if not.
        from .cfglocate import locate_master_cfg_file
        mf = locate_master_cfg_file()
        if mf and mf.is_file():
            from . import io as _io
            _io.make_quiet()
            from .envsetup import emit_envsetup
            emit_envsetup()
        raise SystemExit
    sys.argv[0] = 'sb'
    main( prevent_env_setup_msg = True )

def sbenv_main( args = None):
    if args is None:
        import sys
        args = sys.argv[1:]
    request_help = ( len(args)==1 and args[0] in ('-h','--help') )
    if request_help or not args:
        print("""Usage:

sbenv <program> [args]

Runs <program> within the simplebuild runtime environment. Note that if you wish
to make sure the codebase has been built first (with simplebuild) you should use
sbrun rather than sbenv.
""")
        raise SystemExit(0 if request_help else 111)
        return
    from .envsetup import create_install_env_clone
    run_env = create_install_env_clone()
    from . import utils
    import shlex
    cmd = ' '.join(shlex.quote(e) for e in args)
    ec = utils.system(cmd,env=run_env)
    raise SystemExit(int(ec))

def sbrun_main():
    import sys
    args = sys.argv[1:]
    request_help = ( len(args)==1 and args[0] in ('-h','--help') )
    if request_help or not args:
        print("""Usage:

sbrun <program> [args]

Runs the simplebuild 'sb' command (quietly) and if it finishes successfully,
then proceeds to launch <program> within the simplebuild runtime environment.
""")
        raise SystemExit(0 if request_help else 111)
        return
    from . import frontend
    frontend.simplebuild_main( argv = ['sb',
                                       '--quiet'],
                               prevent_env_setup_msg = True )
    sbenv_main( args )

if __name__=='__main__':
    main()



def produce_build_summary( *, pkgloader, verbose ):
    from . import dirs
    from . import envcfg
    from .utils import path_is_relative_to
    import pathlib
    import os
    from .io import print, print_prefix as prefix

    print('Summary:')
    print()
    def fixpath( p ):
        if envcfg.var.conda_prefix:
            pabs = p.absolute().resolve()
            cp = pathlib.Path(envcfg.var.conda_prefix).absolute().resolve()
            if cp.is_dir() and path_is_relative_to( pabs, cp ):
                return os.path.join('${CONDA_PREFIX}',str(pabs.relative_to(cp)))
        return str(p)
    print('  Projects directory               : %s'%fixpath(dirs.projdir))
    print('  Installation directory           : %s'%fixpath(dirs.installdir))
    print('  Build directory                  : %s'%fixpath(dirs.blddir))

    from . import col
    col_ok = col.ok
    col_bad = col.bad
    col_end = col.end
    #FIXME: Use formatlist module!
    def formatlist(lin,col):
        l=lin[:]
        colbegin = col if col else ''
        colend = col_end if col else ''
        if not l or l==['']:
            return '<none>'
        first=True
        out=''
        while l:
            s=''
            while l and len(s)<40:
                if s:
                    s += ' '
                s += l.pop(0)
            if first:
                out+='%s%s%s'%(colbegin,s,colend)
                first=False
            else:
                out += '\n%s                                      %s%s%s'%(prefix,colbegin,s,colend)
        return out
        #return ' '.join(l)

    pkg_src_info = []
    for basedir in dirs.pkgsearchpath:
        _pkgs_in_basedir = [p for p in pkgloader.pkgs if basedir in p.dirpath.parents]
        pkg_nr = len(_pkgs_in_basedir)
        pkg_enabled = sum(1 for p in _pkgs_in_basedir if p.enabled)
        pkg_src_info.append([basedir, pkg_enabled, (pkg_nr-pkg_enabled) ])

    def pkg_info_str(info):
        p=fixpath(info[0])
        if info[1]+info[2]==0:
            descr='no pkgs'
        else:
            nbuilt = '%s%d%s'%(col_ok,info[1],col_end) if info[1]!=0 else '0'
            if info[2]==0:
                #everything built
                descr=f'{nbuilt} pkgs'
            else:
                nskipped = '%s%d%s'%(col_bad,info[2],col_end)
                descr='%s built, %s skipped'%(nbuilt, nskipped)
        return "%s (%s)"%(p, descr)

    psptxt = '  Package search path              : '
    for idx,info in enumerate(pkg_src_info):
        _= (' '*len(psptxt)) if idx else psptxt
        print(f'{_}{pkg_info_str(info)}')

    nmax = 20
    pkg_enabled = sorted([p.name for p in pkgloader.pkgs if p.enabled])
    n_enabled = len(pkg_enabled)
    lm2='(%i more,'
    limittxt=['...',lm2,'supply','--verbose','to','see','all)']
    if not verbose and n_enabled>nmax:
        limittxt[1] = lm2%(n_enabled-nmax)
        pkg_enabled = pkg_enabled[0:nmax]+limittxt
    pkg_disabled = sorted([p.name for p in pkgloader.pkgs if not p.enabled])
    n_disabled = len(pkg_disabled)
    if not verbose and n_disabled>nmax:
        limittxt[1] = lm2%(n_disabled-nmax)
        pkg_disabled = pkg_disabled[0:nmax]+limittxt
    from . import env
    extdeps_avail = sorted(k for (k,v) in env.env['extdeps'].items() if v['present'])
    extdeps_missing = sorted(k for (k,v) in env.env['extdeps'].items() if not v['present'])

    #Compilers (Fortran is considered optional), CMake, and required externals deps like Python and pybind11:
    reqdep = [('CMake',env.env['system']['general']['cmake_version'])]
    for lang,info in env.env['system']['langs'].items():
        if not info:
            continue
        if info['cased']!='Fortran':
            reqdep += [(info['cased'],info['compiler_version_short'])]
            for dep in info['dep_versions'].split(';'):
                reqdep += [tuple(dep.split('##',1))]
        else:
            assert not info['dep_versions']#If allowed, we would need to print them somewhere

    print('  System                           : %s'%env.env['system']['general']['system'])
    cp=env.env['cmake_printinfo']

    print('  Required dependencies            : %s'%formatlist(['%s[%s]'%(k,v) for k,v in sorted(set(reqdep))],None))
    print('  Optional dependencies present    : %s'%formatlist(['%s[%s]'%(e,env.env['extdeps'][e]['version']) for e in extdeps_avail],
                                                                       col_ok))
    print('  Optional dependencies missing[*] : %s'%formatlist(extdeps_missing,col_bad))
    _pfilter = ( '<none>' if envcfg.var.pkg_filter.fully_open()
                 else "'%s'"%envcfg.var.pkg_filter.as_string() )
    print(f"  Package filter[*]                : {_pfilter}")
    print('  Build mode                       : '
          + f'{envcfg.var.build_mode_summary_string}' )


    print('')
    pkgtxt_en ='%s%i%s package%s built successfully'%(col_ok if n_enabled else '',
                                                      n_enabled,
                                                      col_end if n_enabled else '',
                                                      '' if n_enabled==1 else 's')
    pkgtxt_dis='%s%i%s package%s skipped due to [*]'%(col_bad if n_disabled else '',
                                                      n_disabled,
                                                      col_end if n_disabled else '',
                                                      '' if n_disabled==1 else 's')
    print('  %s : %s'%(pkgtxt_en.ljust(32+(len(col_end)+len(col_ok) if n_enabled else 0)),formatlist(pkg_enabled,col_ok)))
    print('  %s : %s'%(pkgtxt_dis.ljust(32+(len(col_end)+len(col_bad) if n_disabled else 0)),formatlist(pkg_disabled,col_bad)))
    print()
    if cp['other_warnings'] or len(cp['unused_vars'])>0:
        print('%sWARNING%s unspecified warnings from CMake encountered during environment inspection!'%(col_bad,col_end))
        print()
